@@ -3,31 +3,23 @@
 # Workbook - A class for writing the Excel XLSX Workbook file.
 #
 # SPDX-License-Identifier: BSD-2-Clause
-# Copyright 2013-2023, John McNamara, jmcnamara@cpan.org
+#
+# Copyright (c) 2013-2025, John McNamara, jmcnamara@cpan.org
 #
 
 # Standard packages.
-import hashlib
 import operator
 import os
 import re
 import time
-from datetime import datetime
+from datetime import datetime, timezone
 from decimal import Decimal
 from fractions import Fraction
-from struct import unpack
 from warnings import warn
-from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED, LargeZipFile
-
+from zipfile import ZIP_DEFLATED, LargeZipFile, ZipFile, ZipInfo
 
 # Package imports.
 from . import xmlwriter
-from .worksheet import Worksheet
-from .chartsheet import Chartsheet
-from .sharedstrings import SharedStringTable
-from .format import Format
-from .packager import Packager
-from .utility import xl_cell_to_rowcol
 from .chart_area import ChartArea
 from .chart_bar import ChartBar
 from .chart_column import ChartColumn
@@ -37,12 +29,18 @@ from .chart_pie import ChartPie
 from .chart_radar import ChartRadar
 from .chart_scatter import ChartScatter
 from .chart_stock import ChartStock
-from .exceptions import InvalidWorksheetName
-from .exceptions import DuplicateWorksheetName
-from .exceptions import UndefinedImageSize
-from .exceptions import UnsupportedImageFormat
-from .exceptions import FileCreateError
-from .exceptions import FileSizeError
+from .chartsheet import Chartsheet
+from .exceptions import (
+    DuplicateWorksheetName,
+    FileCreateError,
+    FileSizeError,
+    InvalidWorksheetName,
+)
+from .format import Format
+from .packager import Packager
+from .sharedstrings import SharedStringTable
+from .utility import _get_image_properties, xl_cell_to_rowcol
+from .worksheet import Worksheet
 
 
 class Workbook(xmlwriter.XMLwriter):
@@ -68,7 +66,7 @@ class Workbook(xmlwriter.XMLwriter):
         if options is None:
             options = {}
 
-        super(Workbook, self).__init__()
+        super().__init__()
 
         self.filename = filename
 
@@ -121,7 +119,7 @@ class Workbook(xmlwriter.XMLwriter):
         self.custom_colors = []
         self.doc_properties = {}
         self.custom_properties = []
-        self.createtime = datetime.utcnow()
+        self.createtime = datetime.now(timezone.utc)
         self.num_vml_files = 0
         self.num_comment_files = 0
         self.x_window = 240
@@ -131,7 +129,9 @@ class Workbook(xmlwriter.XMLwriter):
         self.tab_ratio = 600
         self.str_table = SharedStringTable()
         self.vba_project = None
-        self.vba_is_stream = False
+        self.vba_project_is_stream = False
+        self.vba_project_signature = None
+        self.vba_project_signature_is_stream = False
         self.vba_codename = None
         self.image_types = {}
         self.images = []
@@ -144,6 +144,11 @@ class Workbook(xmlwriter.XMLwriter):
         self.has_comments = False
         self.read_only = 0
         self.has_metadata = False
+        self.has_embedded_images = False
+        self.has_dynamic_functions = False
+        self.has_embedded_descriptions = False
+        self.embedded_images = EmbeddedImages()
+        self.feature_property_bags = set()
 
         # We can't do 'constant_memory' mode while doing 'in_memory' mode.
         if self.in_memory:
@@ -169,6 +174,7 @@ class Workbook(xmlwriter.XMLwriter):
         return self
 
     def __exit__(self, type, value, traceback):
+        # pylint: disable=redefined-builtin
         """Close workbook when exiting "with" statement."""
         self.close()
 
@@ -248,7 +254,7 @@ class Workbook(xmlwriter.XMLwriter):
         chart_type = options.get("type")
         if chart_type is None:
             warn("Chart type must be defined in add_chart()")
-            return
+            return None
 
         if chart_type == "area":
             chart = ChartArea(options)
@@ -257,20 +263,20 @@ class Workbook(xmlwriter.XMLwriter):
         elif chart_type == "column":
             chart = ChartColumn(options)
         elif chart_type == "doughnut":
-            chart = ChartDoughnut(options)
+            chart = ChartDoughnut()
         elif chart_type == "line":
             chart = ChartLine(options)
         elif chart_type == "pie":
-            chart = ChartPie(options)
+            chart = ChartPie()
         elif chart_type == "radar":
             chart = ChartRadar(options)
         elif chart_type == "scatter":
             chart = ChartScatter(options)
         elif chart_type == "stock":
-            chart = ChartStock(options)
+            chart = ChartStock()
         else:
-            warn("Unknown chart type '%s' in add_chart()" % chart_type)
-            return
+            warn(f"Unknown chart type '{chart_type}' in add_chart()")
+            return None
 
         # Set the embedded chart name if present.
         if "name" in options:
@@ -293,18 +299,49 @@ class Workbook(xmlwriter.XMLwriter):
             is_stream:   vba_project is an in memory byte stream.
 
         Returns:
-            Nothing.
+            0 on success.
 
         """
         if not is_stream and not os.path.exists(vba_project):
-            warn("VBA project binary file '%s' not found." % vba_project)
+            warn(f"VBA project binary file '{vba_project}' not found.")
             return -1
 
         if self.vba_codename is None:
             self.vba_codename = "ThisWorkbook"
 
         self.vba_project = vba_project
-        self.vba_is_stream = is_stream
+        self.vba_project_is_stream = is_stream
+
+        return 0
+
+    def add_signed_vba_project(
+        self, vba_project, signature, project_is_stream=False, signature_is_stream=False
+    ):
+        """
+        Add a vbaProject binary and a vbaProjectSignature binary to the
+        Excel workbook.
+
+        Args:
+            vba_project:           The vbaProject binary file name.
+            signature:             The vbaProjectSignature binary file name.
+            project_is_stream:     vba_project is an in memory byte stream.
+            signature_is_stream:   signature is an in memory byte stream.
+
+        Returns:
+            0 on success.
+
+        """
+        if self.add_vba_project(vba_project, project_is_stream) == -1:
+            return -1
+
+        if not signature_is_stream and not os.path.exists(signature):
+            warn(f"VBA project signature binary file '{signature}' not found.")
+            return -1
+
+        self.vba_project_signature = signature
+        self.vba_project_signature_is_stream = signature_is_stream
+
+        return 0
 
     def close(self):
         """
@@ -317,6 +354,7 @@ class Workbook(xmlwriter.XMLwriter):
             Nothing.
 
         """
+        # pylint: disable=raise-missing-from
         if not self.fileclosed:
             try:
                 self._store_workbook()
@@ -376,7 +414,7 @@ class Workbook(xmlwriter.XMLwriter):
             return
 
         if tab_ratio < 0 or tab_ratio > 100:
-            warn("Tab ratio '%d' outside: 0 <= tab_ratio <= 100" % tab_ratio)
+            warn(f"Tab ratio '{tab_ratio}' outside: 0 <= tab_ratio <= 100")
         else:
             self.tab_ratio = int(tab_ratio * 10)
 
@@ -403,7 +441,7 @@ class Workbook(xmlwriter.XMLwriter):
             property_type: The type of the custom property. Optional.
 
         Returns:
-            Nothing.
+            0 on success.
 
         """
         if name is None or value is None:
@@ -431,18 +469,20 @@ class Workbook(xmlwriter.XMLwriter):
 
         if property_type == "text" and len(value) > 255:
             warn(
-                "Length of 'value' parameter exceeds Excel's limit of 255 "
-                "characters in set_custom_property(): '%s'" % value
+                f"Length of 'value' parameter exceeds Excel's limit of 255 "
+                f"characters in set_custom_property(): '{value}'"
             )
 
         if len(name) > 255:
             warn(
-                "Length of 'name' parameter exceeds Excel's limit of 255 "
-                "characters in set_custom_property(): '%s'" % name
+                f"Length of 'name' parameter exceeds Excel's limit of 255 "
+                f"characters in set_custom_property(): '{name}'"
             )
 
         self.custom_properties.append((name, value, property_type))
 
+        return 0
+
     def set_calc_mode(self, mode, calc_id=None):
         """
         Set the Excel calculation mode for the workbook.
@@ -479,7 +519,7 @@ class Workbook(xmlwriter.XMLwriter):
             formula: The cell or range that the defined name refers to.
 
         Returns:
-            Nothing.
+            0 on success.
 
         """
         sheet_index = None
@@ -490,7 +530,7 @@ class Workbook(xmlwriter.XMLwriter):
             formula = formula.lstrip("=")
 
         # Local defined names are formatted like "Sheet1!name".
-        sheet_parts = re.compile(r"^(.*)!(.*)$")
+        sheet_parts = re.compile(r"^([^!]+)!([^!]+)$")
         match = sheet_parts.match(name)
 
         if match:
@@ -500,7 +540,7 @@ class Workbook(xmlwriter.XMLwriter):
 
             # Warn if the sheet index wasn't found.
             if sheet_index is None:
-                warn("Unknown sheet name '%s' in defined_name()" % sheetname)
+                warn(f"Unknown sheet name '{sheetname}' in defined_name()")
                 return -1
         else:
             # Use -1 to indicate global names.
@@ -510,21 +550,23 @@ class Workbook(xmlwriter.XMLwriter):
         if not re.match(r"^[\w\\][\w\\.]*$", name, re.UNICODE) or re.match(
             r"^\d", name
         ):
-            warn("Invalid Excel characters in defined_name(): '%s'" % name)
+            warn(f"Invalid Excel characters in defined_name(): '{name}'")
             return -1
 
         # Warn if the defined name looks like a cell name.
-        if re.match(r"^[a-zA-Z][a-zA-Z]?[a-dA-D]?[0-9]+$", name):
-            warn("Name looks like a cell name in defined_name(): '%s'" % name)
+        if re.match(r"^[a-zA-Z][a-zA-Z]?[a-dA-D]?\d+$", name):
+            warn(f"Name looks like a cell name in defined_name(): '{name}'")
             return -1
 
         # Warn if the name looks like a R1C1 cell reference.
         if re.match(r"^[rcRC]$", name) or re.match(r"^[rcRC]\d+[rcRC]\d+$", name):
-            warn("Invalid name '%s' like a RC cell ref in defined_name()" % name)
+            warn(f"Invalid name '{name}' like a RC cell ref in defined_name()")
             return -1
 
         self.defined_names.append([name, sheet_index, formula, False])
 
+        return 0
+
     def worksheets(self):
         """
         Return a list of the worksheet objects in the workbook.
@@ -655,6 +697,7 @@ class Workbook(xmlwriter.XMLwriter):
         self._xml_close()
 
     def _store_workbook(self):
+        # pylint: disable=consider-using-with
         # Create the xlsx/zip file.
         try:
             xlsx_file = ZipFile(
@@ -779,10 +822,12 @@ class Workbook(xmlwriter.XMLwriter):
             "nan_inf_to_errors": self.nan_inf_to_errors,
             "default_date_format": self.default_date_format,
             "default_url_format": self.default_url_format,
+            "workbook_add_format": self.add_format,
             "excel2003_style": self.excel2003_style,
             "remove_timezone": self.remove_timezone,
             "max_url_length": self.max_url_length,
             "use_future_functions": self.use_future_functions,
+            "embedded_images": self.embedded_images,
         }
 
         worksheet._initialize(init_data)
@@ -813,19 +858,19 @@ class Workbook(xmlwriter.XMLwriter):
         # Check that sheet sheetname is <= 31. Excel limit.
         if len(sheetname) > 31:
             raise InvalidWorksheetName(
-                "Excel worksheet name '%s' must be <= 31 chars." % sheetname
+                f"Excel worksheet name '{sheetname}' must be <= 31 chars."
             )
 
         # Check that sheetname doesn't contain any invalid characters.
         if invalid_char.search(sheetname):
             raise InvalidWorksheetName(
-                "Invalid Excel character '[]:*?/\\' in sheetname '%s'." % sheetname
+                f"Invalid Excel character '[]:*?/\\' in sheetname '{sheetname}'."
             )
 
         # Check that sheetname doesn't start or end with an apostrophe.
         if sheetname.startswith("'") or sheetname.endswith("'"):
             raise InvalidWorksheetName(
-                'Sheet name cannot start or end with an apostrophe "%s".' % sheetname
+                f'Sheet name cannot start or end with an apostrophe "{sheetname}".'
             )
 
         # Check that the worksheet name doesn't already exist since this is a
@@ -833,7 +878,7 @@ class Workbook(xmlwriter.XMLwriter):
         for worksheet in self.worksheets():
             if sheetname.lower() == worksheet.name.lower():
                 raise DuplicateWorksheetName(
-                    "Sheetname '%s', with case ignored, is already in use." % sheetname
+                    f"Sheetname '{sheetname}', with case ignored, is already in use."
                 )
 
         return sheetname
@@ -939,7 +984,6 @@ class Workbook(xmlwriter.XMLwriter):
         unique_num_formats = {}
         num_formats = []
         index = 164
-        num_format_count = 0
 
         for xf_format in self.xf_formats + self.dxf_formats:
             num_format = xf_format.num_format
@@ -954,11 +998,13 @@ class Workbook(xmlwriter.XMLwriter):
 
                 xf_format.num_format_index = num_format
                 continue
-            elif num_format == "0":
+
+            if num_format == "0":
                 # Number format '0' is indexed as 1 in Excel.
                 xf_format.num_format_index = 1
                 continue
-            elif num_format == "General":
+
+            if num_format == "General":
                 # The 'General' format has an number format index of 0.
                 xf_format.num_format_index = 0
                 continue
@@ -975,7 +1021,6 @@ class Workbook(xmlwriter.XMLwriter):
                 # Only increase font count for XF formats (not DXF formats).
                 if xf_format.xf_index:
                     num_formats.append(num_format)
-                    num_format_count += 1
 
         self.num_formats = num_formats
 
@@ -1077,6 +1122,19 @@ class Workbook(xmlwriter.XMLwriter):
 
         self.fill_count = index
 
+    def _has_feature_property_bags(self):
+        # Check for any format properties that require a feature bag. Currently
+        # this only applies to checkboxes.
+        if not self.feature_property_bags:
+            for xf_format in self.formats:
+                if xf_format.checkbox:
+                    self.feature_property_bags.add("XFComplements")
+
+                if xf_format.dxf_index is not None and xf_format.checkbox:
+                    self.feature_property_bags.add("DXFComplements")
+
+        return self.feature_property_bags
+
     def _prepare_defined_names(self):
         # Iterate through the worksheets and store any defined names in
         # addition to any user defined names. Stores the defined names
@@ -1149,13 +1207,21 @@ class Workbook(xmlwriter.XMLwriter):
     def _prepare_drawings(self):
         # Iterate through the worksheets and set up chart and image drawings.
         chart_ref_id = 0
-        image_ref_id = 0
         ref_id = 0
         drawing_id = 0
         image_ids = {}
         header_image_ids = {}
         background_ids = {}
 
+        # Store the image types for any embedded images.
+        for image_data in self.embedded_images.images:
+            image_type = image_data[1]
+            self.image_types[image_type] = True
+            if image_data[3]:
+                self.has_embedded_descriptions = True
+
+        image_ref_id = len(self.embedded_images.images)
+
         for sheet in self.worksheets():
             chart_count = len(sheet.charts)
             image_count = len(sheet.images)
@@ -1192,13 +1258,15 @@ class Workbook(xmlwriter.XMLwriter):
 
                 (
                     image_type,
-                    width,
-                    height,
-                    name,
-                    x_dpi,
-                    y_dpi,
+                    _,
+                    _,
+                    _,
+                    _,
+                    _,
                     digest,
-                ) = self._get_image_properties(filename, image_data)
+                ) = _get_image_properties(filename, image_data)
+
+                self.image_types[image_type] = True
 
                 if digest in background_ids:
                     ref_id = background_ids[digest]
@@ -1222,7 +1290,9 @@ class Workbook(xmlwriter.XMLwriter):
                     x_dpi,
                     y_dpi,
                     digest,
-                ) = self._get_image_properties(filename, image_data)
+                ) = _get_image_properties(filename, image_data)
+
+                self.image_types[image_type] = True
 
                 if digest in image_ids:
                     ref_id = image_ids[digest]
@@ -1268,7 +1338,9 @@ class Workbook(xmlwriter.XMLwriter):
                     x_dpi,
                     y_dpi,
                     digest,
-                ) = self._get_image_properties(filename, image_data)
+                ) = _get_image_properties(filename, image_data)
+
+                self.image_types[image_type] = True
 
                 if digest in header_image_ids:
                     ref_id = header_image_ids[digest]
@@ -1304,7 +1376,9 @@ class Workbook(xmlwriter.XMLwriter):
                     x_dpi,
                     y_dpi,
                     digest,
-                ) = self._get_image_properties(filename, image_data)
+                ) = _get_image_properties(filename, image_data)
+
+                self.image_types[image_type] = True
 
                 if digest in header_image_ids:
                     ref_id = header_image_ids[digest]
@@ -1315,7 +1389,7 @@ class Workbook(xmlwriter.XMLwriter):
                     self.images.append([filename, image_type, image_data])
 
                 sheet._prepare_header_image(
-                    image_ref_id,
+                    ref_id,
                     width,
                     height,
                     name,
@@ -1341,246 +1415,6 @@ class Workbook(xmlwriter.XMLwriter):
 
         self.drawing_count = drawing_id
 
-    def _get_image_properties(self, filename, image_data):
-        # Extract dimension information from the image file.
-        height = 0
-        width = 0
-        x_dpi = 96
-        y_dpi = 96
-
-        if not image_data:
-            # Open the image file and read in the data.
-            fh = open(filename, "rb")
-            data = fh.read()
-        else:
-            # Read the image data from the user supplied byte stream.
-            data = image_data.getvalue()
-
-        digest = hashlib.sha256(data).hexdigest()
-
-        # Get the image filename without the path.
-        image_name = os.path.basename(filename)
-
-        # Look for some common image file markers.
-        marker1 = unpack("3s", data[1:4])[0]
-        marker2 = unpack(">H", data[:2])[0]
-        marker3 = unpack("2s", data[:2])[0]
-        marker4 = unpack("<L", data[:4])[0]
-        marker5 = (unpack("4s", data[40:44]))[0]
-        marker6 = unpack("4s", data[:4])[0]
-
-        png_marker = b"PNG"
-        bmp_marker = b"BM"
-        emf_marker = b" EMF"
-        gif_marker = b"GIF8"
-
-        if marker1 == png_marker:
-            self.image_types["png"] = True
-            (image_type, width, height, x_dpi, y_dpi) = self._process_png(data)
-
-        elif marker2 == 0xFFD8:
-            self.image_types["jpeg"] = True
-            (image_type, width, height, x_dpi, y_dpi) = self._process_jpg(data)
-
-        elif marker3 == bmp_marker:
-            self.image_types["bmp"] = True
-            (image_type, width, height) = self._process_bmp(data)
-
-        elif marker4 == 0x9AC6CDD7:
-            self.image_types["wmf"] = True
-            (image_type, width, height, x_dpi, y_dpi) = self._process_wmf(data)
-
-        elif marker4 == 1 and marker5 == emf_marker:
-            self.image_types["emf"] = True
-            (image_type, width, height, x_dpi, y_dpi) = self._process_emf(data)
-
-        elif marker6 == gif_marker:
-            self.image_types["gif"] = True
-            (image_type, width, height, x_dpi, y_dpi) = self._process_gif(data)
-
-        else:
-            raise UnsupportedImageFormat(
-                "%s: Unknown or unsupported image file format." % filename
-            )
-
-        # Check that we found the required data.
-        if not height or not width:
-            raise UndefinedImageSize("%s: no size data found in image file." % filename)
-
-        if not image_data:
-            fh.close()
-
-        # Set a default dpi for images with 0 dpi.
-        if x_dpi == 0:
-            x_dpi = 96
-        if y_dpi == 0:
-            y_dpi = 96
-
-        return image_type, width, height, image_name, x_dpi, y_dpi, digest
-
-    def _process_png(self, data):
-        # Extract width and height information from a PNG file.
-        offset = 8
-        data_length = len(data)
-        end_marker = False
-        width = 0
-        height = 0
-        x_dpi = 96
-        y_dpi = 96
-
-        # Search through the image data to read the height and width in the
-        # IHDR element. Also read the DPI in the pHYs element.
-        while not end_marker and offset < data_length:
-            length = unpack(">I", data[offset + 0 : offset + 4])[0]
-            marker = unpack("4s", data[offset + 4 : offset + 8])[0]
-
-            # Read the image dimensions.
-            if marker == b"IHDR":
-                width = unpack(">I", data[offset + 8 : offset + 12])[0]
-                height = unpack(">I", data[offset + 12 : offset + 16])[0]
-
-            # Read the image DPI.
-            if marker == b"pHYs":
-                x_density = unpack(">I", data[offset + 8 : offset + 12])[0]
-                y_density = unpack(">I", data[offset + 12 : offset + 16])[0]
-                units = unpack("b", data[offset + 16 : offset + 17])[0]
-
-                if units == 1:
-                    x_dpi = x_density * 0.0254
-                    y_dpi = y_density * 0.0254
-
-            if marker == b"IEND":
-                end_marker = True
-                continue
-
-            offset = offset + length + 12
-
-        return "png", width, height, x_dpi, y_dpi
-
-    def _process_jpg(self, data):
-        # Extract width and height information from a JPEG file.
-        offset = 2
-        data_length = len(data)
-        end_marker = False
-        width = 0
-        height = 0
-        x_dpi = 96
-        y_dpi = 96
-
-        # Search through the image data to read the JPEG markers.
-        while not end_marker and offset < data_length:
-            marker = unpack(">H", data[offset + 0 : offset + 2])[0]
-            length = unpack(">H", data[offset + 2 : offset + 4])[0]
-
-            # Read the height and width in the 0xFFCn elements (except C4, C8
-            # and CC which aren't SOF markers).
-            if (
-                (marker & 0xFFF0) == 0xFFC0
-                and marker != 0xFFC4
-                and marker != 0xFFC8
-                and marker != 0xFFCC
-            ):
-                height = unpack(">H", data[offset + 5 : offset + 7])[0]
-                width = unpack(">H", data[offset + 7 : offset + 9])[0]
-
-            # Read the DPI in the 0xFFE0 element.
-            if marker == 0xFFE0:
-                units = unpack("b", data[offset + 11 : offset + 12])[0]
-                x_density = unpack(">H", data[offset + 12 : offset + 14])[0]
-                y_density = unpack(">H", data[offset + 14 : offset + 16])[0]
-
-                if units == 1:
-                    x_dpi = x_density
-                    y_dpi = y_density
-
-                if units == 2:
-                    x_dpi = x_density * 2.54
-                    y_dpi = y_density * 2.54
-
-                # Workaround for incorrect dpi.
-                if x_dpi == 1:
-                    x_dpi = 96
-                if y_dpi == 1:
-                    y_dpi = 96
-
-            if marker == 0xFFDA:
-                end_marker = True
-                continue
-
-            offset = offset + length + 2
-
-        return "jpeg", width, height, x_dpi, y_dpi
-
-    def _process_gif(self, data):
-        # Extract width and height information from a GIF file.
-        x_dpi = 96
-        y_dpi = 96
-
-        width = unpack("<h", data[6:8])[0]
-        height = unpack("<h", data[8:10])[0]
-
-        return "gif", width, height, x_dpi, y_dpi
-
-    def _process_bmp(self, data):
-        # Extract width and height information from a BMP file.
-        width = unpack("<L", data[18:22])[0]
-        height = unpack("<L", data[22:26])[0]
-        return "bmp", width, height
-
-    def _process_wmf(self, data):
-        # Extract width and height information from a WMF file.
-        x_dpi = 96
-        y_dpi = 96
-
-        # Read the bounding box, measured in logical units.
-        x1 = unpack("<h", data[6:8])[0]
-        y1 = unpack("<h", data[8:10])[0]
-        x2 = unpack("<h", data[10:12])[0]
-        y2 = unpack("<h", data[12:14])[0]
-
-        # Read the number of logical units per inch. Used to scale the image.
-        inch = unpack("<H", data[14:16])[0]
-
-        # Convert to rendered height and width.
-        width = float((x2 - x1) * x_dpi) / inch
-        height = float((y2 - y1) * y_dpi) / inch
-
-        return "wmf", width, height, x_dpi, y_dpi
-
-    def _process_emf(self, data):
-        # Extract width and height information from a EMF file.
-
-        # Read the bounding box, measured in logical units.
-        bound_x1 = unpack("<l", data[8:12])[0]
-        bound_y1 = unpack("<l", data[12:16])[0]
-        bound_x2 = unpack("<l", data[16:20])[0]
-        bound_y2 = unpack("<l", data[20:24])[0]
-
-        # Convert the bounds to width and height.
-        width = bound_x2 - bound_x1
-        height = bound_y2 - bound_y1
-
-        # Read the rectangular frame in units of 0.01mm.
-        frame_x1 = unpack("<l", data[24:28])[0]
-        frame_y1 = unpack("<l", data[28:32])[0]
-        frame_x2 = unpack("<l", data[32:36])[0]
-        frame_y2 = unpack("<l", data[36:40])[0]
-
-        # Convert the frame bounds to mm width and height.
-        width_mm = 0.01 * (frame_x2 - frame_x1)
-        height_mm = 0.01 * (frame_y2 - frame_y1)
-
-        # Get the dpi based on the logical size.
-        x_dpi = width * 25.4 / width_mm
-        y_dpi = height * 25.4 / height_mm
-
-        # This is to match Excel's calculation. It is probably to account for
-        # the fact that the bounding box is inclusive-inclusive. Or a bug.
-        width += 1
-        height += 1
-
-        return "emf", width, height, x_dpi, y_dpi
-
     def _extract_named_ranges(self, defined_names):
         # Extract the named ranges from the sorted list of defined names.
         # These are used in the App.xml file.
@@ -1616,8 +1450,8 @@ class Workbook(xmlwriter.XMLwriter):
 
         if sheetname in self.sheetnames:
             return self.sheetnames[sheetname].index
-        else:
-            return None
+
+        return None
 
     def _prepare_vml(self):
         # Iterate through the worksheets and set up the VML objects.
@@ -1675,9 +1509,13 @@ class Workbook(xmlwriter.XMLwriter):
 
     def _prepare_metadata(self):
         # Set the metadata rel link.
+        self.has_embedded_images = self.embedded_images.has_images()
+        self.has_metadata = self.has_embedded_images
+
         for sheet in self.worksheets():
             if sheet.has_dynamic_arrays:
                 self.has_metadata = True
+                self.has_dynamic_functions = True
 
     def _add_chart_data(self):
         # Add "cached" data to charts to provide the numCache and strCache
@@ -1731,8 +1569,8 @@ class Workbook(xmlwriter.XMLwriter):
                 # in a chart series formula.
                 if sheetname not in worksheets:
                     warn(
-                        "Unknown worksheet reference '%s' in range "
-                        "'%s' passed to add_series()" % (sheetname, c_range)
+                        f"Unknown worksheet reference '{sheetname}' in range "
+                        f"'{c_range}' passed to add_series()"
                     )
                     chart.formula_data[r_id] = []
                     seen_ranges[c_range] = []
@@ -1744,8 +1582,6 @@ class Workbook(xmlwriter.XMLwriter):
                 # Get the data from the worksheet table.
                 data = worksheet._get_range_data(*cells)
 
-                # TODO. Handle SST string ids if required.
-
                 # Add the data to the chart.
                 chart.formula_data[r_id] = data
 
@@ -1908,8 +1744,10 @@ class Workbook(xmlwriter.XMLwriter):
             ("sheetId", sheet_id),
         ]
 
-        if hidden:
+        if hidden == 1:
             attributes.append(("state", "hidden"))
+        elif hidden == 2:
+            attributes.append(("state", "veryHidden"))
 
         attributes.append(("r:id", "rId" + str(sheet_id)))
 
@@ -1960,7 +1798,7 @@ class Workbook(xmlwriter.XMLwriter):
 
 
 # A metadata class to share data between worksheets.
-class WorksheetMeta(object):
+class WorksheetMeta:
     """
     A class to track worksheets data such as the active sheet and the
     first sheet.
@@ -1970,3 +1808,49 @@ class WorksheetMeta(object):
     def __init__(self):
         self.activesheet = 0
         self.firstsheet = 0
+
+
+# A helper class to share embedded images between worksheets.
+class EmbeddedImages:
+    """
+    A class to track duplicate embedded images between worksheets.
+
+    """
+
+    def __init__(self):
+        self.images = []
+        self.image_indexes = {}
+
+    def get_image_index(self, image, digest):
+        """
+        Get the index of an embedded image.
+
+        Args:
+            image: The image to lookup.
+            digest: The digest of the image.
+
+        Returns:
+            The image index.
+
+        """
+        image_index = self.image_indexes.get(digest)
+
+        if image_index is None:
+            self.images.append(image)
+            image_index = len(self.images)
+            self.image_indexes[digest] = image_index
+
+        return image_index
+
+    def has_images(self):
+        """
+        Check if the worksheet has embedded images.
+
+        Args:
+            None.
+
+        Returns:
+            Boolean.
+
+        """
+        return len(self.images) > 0
